// Order.java

import swarm.Globals;
import swarm.defobj.Zone;
import swarm.objectbase.SwarmObjectImpl;
import swarm.Selector;

import swarm.collections.ListImpl;
import swarm.collections.ListIndex;

/**
 * The form containing our Order metaphoric object
 *
 * @author Pietro Terna (with Paolo Pelligra and Dario Landini contibutions)
 */
public class Order  extends SwarmObjectImpl implements Cloneable{

    /** the number of steps in the product recipe */
    public int stepNumber;
    /** the 'name' of the recipe (may be used by OrderDistiller, as a way
     *  to identify recipes */
    public String recipeName="noName";
    /** the layer to which the order belongs; the layer has to be set by
     *  OrderGenerator (for test purposes) and must of all by the 
     *  OrderDistiller; in this case, the information must be retrieved from
     *  the recipeData/orderSequence.xls file
     *
     *  a layer is a period of time or a set of qualitative conditions that 
     *  introduce differentiations into the orders; e.g two collections in 
     *  fashion production, with the same technical description (recipe) and 
     *  different qualitative results */
    public int orderLayer=0; 
    /** the number of this order */
    public int orderNumber;
    /** multiplicity factor, usually = 1; > 1 in stand alone batch 
     *  productions */
    public int multiplicity = 1;
    /** the day in which the order as been generated */
    public int startingTimeUnit;
    /** integer vector, containing the recipe steps */
    public int[] orderRecipe;
    /** a temporary vector to redefine the dimension of orderRecipe
     *  if necessary (using  "p" or "c" or "||" or "&&" steps) */ 
    public int[] orderRecipeTmp;
    /** integer vector, recording the performed phases */
    public int[] orderState;
    /** the sequential number of the recipe referring to which
     * a news will be sent */
    public int numberOfTheStepForWhichANewsIsSent;
    /** the fixed and variable costs of the order */
    public float cumulatedFixedCost, cumulatedVariableCost;
    /** the sum of variable and fixed costs of the order */
    public float orderTotalCosts = 0;

    /** a procurement specification set */
    public ProcurementSpecificationSet aProcurementSpecificationSet;
    /** the list of the procurement specification sets */
    public ListImpl procurementSpecificationSetList;

    /** compitational specification set */
    public ComputationalSpecificationSet aComputationalSpecificationSet;
    /** the list of the computational specification set */
    public ListImpl computationalSpecificationSetList;
    /** the vector of memory matrix addresses to be set in a
     *  computational specification set */
    public Object [] matrixAddress;

    /** a stand alone batch specification sets */
    public StandAloneBatchSpecificationSet aStandAloneBatchSpecificationSet;
    /** the list of stand alone batch specifications */
    public ListImpl standAloneBatchSpecificationSetList;

    /** a sequential batch specification set */
    public SequentialBatchSpecificationSet aSequentialBatchSpecificationSet;
    /** the list of sequential batch specifications */
    public ListImpl sequentialBatchSpecificationSetList;

    /** an 'or' order specification set */
    public OrSpecificationSet anOrSpecificationSet;
    /** the list of 'or' specifications */
    public ListImpl orSpecificationSetList;
    /** the row used in the memory matrix invoked by an 'or' process using
     *  orCriterion 5 (look at Or.java) */
    public int orRow; 

    // iterators
    ListIndex procurementSpecificationSetListIndex;
    ListIndex computationalSpecificationSetListIndex;
    ListIndex procuredItemListIndex;
    ListIndex standAloneBatchSpecificationSetListIndex;
    ListIndex sequentialBatchSpecificationSetListIndex;
    ListIndex orSpecificationSetListIndex;

     /** the list of procured items */
    public ListImpl procuredItemList;

    /** the model */
    VEFrameModelSwarm vEFrameModelSwarm;

    /** the end unit list */
    public ListImpl endUnitList;
    /** its iterator */
    public ListIndex endUnitListIndex;

    /** inSequentialBatch is true if the order is found to be included in a 
     *  sequential batch */
    public boolean inSequentialBatch;

    /** indexes used in a common way with the arrangeForProcurements() 
	internal method */
    public int i, j, jj, k;

    /** incoming recipe structure; we make a local copy of it, from r0 which
     *  is in the constructor paramenter list, to r */
    public int[] r;

  /**
   * the first constructor for Order, for original object creation
   */
  public Order (Zone aZone, int n, int sd, int sn, int [] r0, 
		VEFrameModelSwarm model, ListImpl eul) {

     // Call the constructor for the Order parent class
     super(aZone);

     r = new int[r0.length];
     for (i=0;i<r0.length;i++)r[i]=r0[i];

     orderNumber=n;
     stepNumber=sn; // it contains also the length of "p" or "c" " or "||" 
                    // or "&&"sequences
     startingTimeUnit=sd;
     orderRecipe= new int[stepNumber];
     orderState = new int[stepNumber];
     // the model
     vEFrameModelSwarm = model;
     // the end unit list
     endUnitList=eul;
     // its iterator
     endUnitListIndex = endUnitList.listBegin(getZone());

     inSequentialBatch=false;

     int node;
     boolean isAnEndUnit, orSequence;

     procurementSpecificationSetList = new ListImpl(getZone());
     procurementSpecificationSetListIndex = 
	    procurementSpecificationSetList.listBegin(getZone());

     procuredItemList = new ListImpl(getZone());
     procuredItemListIndex = 
	 procuredItemList.listBegin(getZone());

     computationalSpecificationSetList = new ListImpl(getZone());
     computationalSpecificationSetListIndex = 
	    computationalSpecificationSetList.listBegin(getZone());

     matrixAddress = new Object[vEFrameModelSwarm.
			       getTotalMemoryMatrixNumber()];

     standAloneBatchSpecificationSetList = new ListImpl(getZone());
     standAloneBatchSpecificationSetListIndex =
	 standAloneBatchSpecificationSetList.listBegin(getZone());

     sequentialBatchSpecificationSetList = new ListImpl(getZone());
     sequentialBatchSpecificationSetListIndex = 
	 sequentialBatchSpecificationSetList.listBegin(getZone());

     orSpecificationSetList = new ListImpl(getZone());
     orSpecificationSetListIndex = 
	 orSpecificationSetList.listBegin(getZone());

     i=0; j=0;
     // copying the recipe, analysing and using "p" o "c"" or "||" or "&&"
     // or "batch" sequences
     while (j<stepNumber)
	 {
	     if (r[j]==0){
		 System.out.
		     println("Internal error in order generation: " + 
			     "look at the constructor of Order class.");
		 MyExit.exit(1);
	     }

	     if (r[j]>0) orderRecipe[i]=r[j];

	     // "p" or "c"" or "||" or "&&"  or "batch" cases
	     if (r[j]<0){
		 if(r[j]==-1)arrangeForProcurements(); // "p" case

		 else if(r[j]==-2){ // "stand alone batch order" case
		     jj=j;
		     j=stepNumber;
		     setMultiplicity(r[jj+2]);
		     orderRecipe[i++] = r[jj+3];

		     // next unit has to be and end unit
		     isAnEndUnit = false;
		     endUnitListIndex.setOffset(0);
		     for(k=0;k<endUnitList.getCount();k++){
			 if(r[jj+4]==((EndUnit) endUnitListIndex.get()).
			    getUnitNumber())isAnEndUnit = true;
			 endUnitListIndex.next();
		     }
		     if(isAnEndUnit == false)
			 {
			     System.out.println(
			     "Syntax error in Order # " + getOrderNumber());
			     System.out.println("An endUnit has to follow " + 
			      "immediately after the stand alone batch " +
			      "process");
			     MyExit.exit(1);
			 }

		     orderRecipe[i]   = r[jj+4];

		     // making a standAloneBatchSpecificationSet object, linked
		     // at the position '0' of the recipe
		     if (jj != 0){
			 System.out.println("Syntax or internal error: " +
					    "a stand alone batch process has "
					    + "to be the first step in a " +
					    "recipe");
			 MyExit.exit(1);
		     }
		     aStandAloneBatchSpecificationSet = new 
			 StandAloneBatchSpecificationSet(getZone(), 
						     orderNumber, 0);

		     standAloneBatchSpecificationSetList.
			 addLast(aStandAloneBatchSpecificationSet);

		     // setting specifications (production time in ticks) into 
		     // aStandAloneBatchSpecificazioneSet
		     aStandAloneBatchSpecificationSet.
			 setSpecifications(r[jj+1]);
		     }

		 else if(r[j]==-3){ // "sequential batch order" case
		     jj=j;
		     j += 3;
		     orderRecipe[i]=r[j];

		     // making a sequentialBatchSpecificationSet object, linked
		     // at the position 'jj' of the recipe

		     aSequentialBatchSpecificationSet = new 
			 SequentialBatchSpecificationSet(getZone(), 
						     orderNumber, i);
		     sequentialBatchSpecificationSetList.
			 addLast(aSequentialBatchSpecificationSet);

		     // setting specifications (quantity in batch;
		     // production time in ticks) into 
		     // aSequentialBatchSpecificazioneSet
		     aSequentialBatchSpecificationSet.
			 setSpecifications(r[jj+1], r[jj+2]);

		 }

		 else if (r[j]==-10){ // "or"
		     // preparing internal recipe and checking the sequence 
		     // of the nodes opening the various branches of an 'or'
		     // sequence
		     orSequence=true;
		     node=1;
		     anOrSpecificationSet = new 
			 OrSpecificationSet(getZone(), orderNumber, i);
		     orSpecificationSetList.addLast(anOrSpecificationSet);

		     anOrSpecificationSet.setNodePosition(node, i);

		     // the first node (it is numbered 1 or 10xx,
		     // where xx is a number un the range 0-99)
		     if(r[j+1] != node && ! (r[j+1]>=1000 && r[j+1] <=1099))
			 {
			     System.out.println("In order "+orderNumber+
				      " found 'or' branch "+r[j+1]+", "+
			              "while waiting for node 1 or 10xx");
			     MyExit.exit(1);
			 }

		     // orRow, i.e. the row used in the memory matrix 
		     // invoked by an 'or' process using orCriterion 5 
		     // (look at Or.java)
		     orRow=0;
		     if(r[j+1]>=1000 && r[j+1]<=1099){
			 orRow=r[j+1]-1000;
			 r[j+1]=1;
		     }

		     node++;
		     j+=2;
		     while(orSequence){
			 if(r[j]==-10 && r[j+1] != node && r[j+1]!=0){
			     System.out.println("In order "+orderNumber+
						" found 'or' branch "+r[j+1]+
						", while waiting for node: "+
						node);
			     MyExit.exit(1);
			 }

			 else if (r[j]==-10 && r[j+1]==node){
			     anOrSpecificationSet.setNodePosition(node, i);
			     node++;
			     j+=2;
			 }
			 else if (r[j]==-10 && r[j+1]==0){
			     orSequence=false;
			     anOrSpecificationSet.
				 setNodePosition(node, i); // the end point

			     anOrSpecificationSet.setNodeNumber(--node);
			     anOrSpecificationSet.setOrRow(orRow);

			     // we are exiting from the 'or' cicle:
			     // i and j will be incresed in the general
			     // while block, so here we go back with i and
			     // increase j only of one step
			     i--;
			     j++;}

			 else if(r[j]==-1)arrangeForProcurements(); // "p" case
			 else if(r[j]>=-1999 && r[j]<=-1001)
			     arrangeForComputation();               // "c" case

			 else {
			     orderRecipe[i]=r[j];
			     i++;
			     j++;
			 }
		     }

		 }

		 else if(r[j]>=-1999 && r[j]<=-1001)
		     arrangeForComputation(); // "c" case

		 else{
		     System.out.
			  println("Internal error analysing the order: " + 
				  "look at the constructor of Order class.");
		 MyExit.exit(1);}
	     }

	     i++; j++;

	 }

     // i is now the internal length of vectors, so we redefine the 
     // stepNumber
     stepNumber=i;

     // redefine orderRecipe vector
     orderRecipeTmp = new int[stepNumber];
     System.arraycopy(orderRecipe,0,orderRecipeTmp,0,stepNumber);
     orderRecipe=orderRecipeTmp;

     // announcing the order presence to the console
     if(StartVEFrame.verbose)
	 System.out.println("New order (#" + orderNumber + ")");

     // setting the state vector (0=empty; not zero=done)
     if(StartVEFrame.verbose)System.out.print("orderState  vector ");
     for (i=1;i<=stepNumber;i++) {
	 orderState[i-1]=0;
	 if(StartVEFrame.verbose)System.out.print(orderState[i-1] + " ");
     }
     if(StartVEFrame.verbose)System.out.println();

     // announcing the recipe vector
     if(StartVEFrame.verbose)
	 {
	 System.out.print("orderRecipe vector ");
	 for (i=1;i<=stepNumber;i++)
	     System.out.print(orderRecipe[i-1] + " ");
	 System.out.println();
     }
  }

  /**
   * the second constructor for Order, to clone an order <br>
   *
   * the cloning operation anyway is only partial
   *
   */
    public Order (Zone aZone, int originalOrderRecipeLength,
		  int[] originalOrderRecipe,
		  int originalOrderStateLength,
		  int[] originalOrderState, VEFrameModelSwarm model,
		  ListIndex originalProcurementSpecificationSetListIndex,
		  ListIndex originalProcuredItemListIndex,
		  ListIndex originalComputationalSpecificationSetListIndex,
		  ListIndex originalStandAloneBatchSpecificationSetListIndex,
		  ListIndex originalSequentialBatchSpecificationSetListIndex,
		  ListIndex originalOrSpecificationSetListIndex,
		  boolean originalInSequentialBatch)
    {
	// Call the constructor for the Order parent class
	super(aZone);

	int i;

	// copy existing vectors
	orderRecipe = new int[originalOrderRecipeLength];
	System.arraycopy(originalOrderRecipe,0,
			 orderRecipe,0,originalOrderRecipeLength);

	orderState = new int[originalOrderStateLength];
	System.arraycopy(originalOrderState,0,
			 orderState,0,originalOrderStateLength);

	// the model
	vEFrameModelSwarm = model;

	// the procurementSpecificationSetList and its contents
	procurementSpecificationSetList = new ListImpl(getZone());
	if(((ListImpl)originalProcurementSpecificationSetListIndex.
	    getCollection()).getCount()>0){
	    originalProcurementSpecificationSetListIndex.setOffset(0);
	    for(i=0;i<((ListImpl)originalProcurementSpecificationSetListIndex.
		       getCollection()).getCount();i++){
		procurementSpecificationSetList.addLast(
			   originalProcurementSpecificationSetListIndex.get());
		originalProcurementSpecificationSetListIndex.next();
	    }
	}

	// the procuredItemList
	procuredItemList = new ListImpl(getZone());
	if(((ListImpl)originalProcuredItemListIndex.getCollection()).
	   getCount()>0){
	    originalProcuredItemListIndex.setOffset(0);
	    for(i=0;i<((ListImpl)originalProcuredItemListIndex.
		       getCollection()).getCount();i++){
		procuredItemList.addLast(
					 originalProcuredItemListIndex.get());
		originalProcuredItemListIndex.next();
	    }
	}

	// the computationalSpecificationSetList and its contents
	computationalSpecificationSetList = new ListImpl(getZone());
	if(((ListImpl)originalComputationalSpecificationSetListIndex.
	    getCollection()).getCount()>0){
	    originalComputationalSpecificationSetListIndex.setOffset(0);
	    for(i=0;i<
                     ((ListImpl)originalComputationalSpecificationSetListIndex.
		       getCollection()).getCount();i++){
		computationalSpecificationSetList.addLast(
			 originalComputationalSpecificationSetListIndex.get());
		originalComputationalSpecificationSetListIndex.next();
	    }
	}

	// the standAloneBatchSpecificationSetList and its contents
	standAloneBatchSpecificationSetList = new ListImpl(getZone());
	if(((ListImpl)originalStandAloneBatchSpecificationSetListIndex.
	    getCollection()).getCount()>0){
	    originalStandAloneBatchSpecificationSetListIndex.setOffset(0);
	    for(i=0;i<
		 ((ListImpl)originalStandAloneBatchSpecificationSetListIndex.
		  getCollection()).getCount();i++){
		standAloneBatchSpecificationSetList.addLast(
		      originalStandAloneBatchSpecificationSetListIndex.get());
		originalStandAloneBatchSpecificationSetListIndex.next();
	    }
	}

	// the sequentialBatchSpecificationSetList and its contents
	sequentialBatchSpecificationSetList = new ListImpl(getZone());
	if(((ListImpl)originalSequentialBatchSpecificationSetListIndex.
	    getCollection()).getCount()>0){
	    originalSequentialBatchSpecificationSetListIndex.setOffset(0);
	    for(i=0;i<
		 ((ListImpl)originalSequentialBatchSpecificationSetListIndex.
		  getCollection()).getCount();i++){
		sequentialBatchSpecificationSetList.addLast(
		       originalSequentialBatchSpecificationSetListIndex.get());
		originalSequentialBatchSpecificationSetListIndex.next();
	    }
	}

	// the OrSpecificationSetList and its contents
	orSpecificationSetList = new ListImpl(getZone());
	if(((ListImpl)originalOrSpecificationSetListIndex.
	    getCollection()).getCount()>0){
	    originalOrSpecificationSetListIndex.setOffset(0);
	    for(i=0;i<((ListImpl)originalOrSpecificationSetListIndex.
		       getCollection()).getCount();i++){
		orSpecificationSetList.addLast(
				 originalOrSpecificationSetListIndex.get());
		originalOrSpecificationSetListIndex.next();
	    }
	}

	// in batch
	inSequentialBatch=originalInSequentialBatch;

  }

  /**
   * production next step in Order (it is in the orderRecipe vector,
   * at the same index value in which we find the first 0 in orderState
   * vector)
   */
    public int getNextStep () {

	int i, nextStep;

	nextStep=0;
	for (i=0;i< stepNumber && nextStep==0;i++)
	    if (orderState[i]==0)nextStep=getOrderRecipe(i);

	// if no 0 position can be found the order is finished
	// in this case we return 0

	if(nextStep>=1000000000)nextStep-=1000000000;
	return nextStep;
  }

    /** 
     *  check if next step in order is a zero time step
     */
    public boolean checkZeroTimeProduction() {

	int i, nextStep;

	nextStep=0;
	for (i=0;i< stepNumber && nextStep==0;i++)
	    if (orderState[i]==0)nextStep=getOrderRecipe(i);

	// if no 0 position can be found the order is finished
	// in this case nextStep = 0

	if(nextStep>=1000000000)return true;
	else return false;
    }

    /**
     * check if next step in Order uses a procurement specification set
     */
    public boolean checkIfNextStepUsesAProcurementSpecificationSet() {
	
	if(getPendingProcurementSpecificationSet() != null)
	    return true;
	else
	    return false;
  }

    /**
     * check if next step in Order uses a computational specification set
     */
    public boolean checkIfNextStepUsesAComputationalSpecificationSet() {
	
	if(getPendingComputationalSpecificationSet() != null)
	    return true;
	else
	    return false;
  }


 /**
   * check if next step in Order uses a stand alone batch specification set
   */
    public boolean checkIfNextStepUsesAStandAloneBatchSpecificationSet() {

	if(getPendingStandAloneBatchSpecificationSet() != null)
	    return true;
	else
	    return false;
  }

 /**
   * check if next step in Order uses a sequential batch specification set
   */
    public boolean checkIfNextStepUsesASequentialBatchSpecificationSet() {

	// if inSequentialBatch is true, the order is being reassigned to a 
	// unit by its sequential bactch assembler
	if(getPendingSequentialBatchSpecificationSet() != null && 
	   ! inSequentialBatch)
	    return true;
	else
	    return false;
  }

    /**
     * check if next step in Order uses a 'or' specification set
     */
    public boolean checkIfNextStepUsesAnOrSpecificationSet() {

	// if 'or' is true, assigningTool has to send the order to Or.java
	// to apply orCriterion to chose the active 'or' branch
	if(getPendingOrSpecificationSet() != null)
	    return true;
	else
	    return false;
  }


    /**
     * 
     */
    public void setOrderLayer(int la){
	orderLayer=la;
    }

    /**
     * 
     */
    public int getOrderLayer(){
	return orderLayer;
    }

    /**
     * Done step in Order
     */
    public void setDoneStep (int u) {

	int i, done;

	done=0;
	for (i=0;i<= stepNumber && done==0;i++)
	    if (orderState[i]==0)
		{orderState[i]=u;
		done=1;
		}
    }

  /**
   * returning the orderNumber of this order
   */
    public int getOrderNumber () {
	return orderNumber;
  }

  /**
   * returning the stepNumber
   */
    public int getStepNumber () {

	return stepNumber;
  }

  /**
   * returning the startingTimeUnit
   */
    public int getStartingTimeUnit() {

	return startingTimeUnit;
  }

  /**
   * returning how many steps remain to terminate the production
   */
    public int getHowManyStepsToBeDone () {

	int i, stepsToBeDone;
	for (i=0;i<stepNumber && orderState[i]!=0;i++); 
                              // to find the first position
	                      // where orderState[i] == 0

	// NB in this way we deliberately ignore the presence of 1 sequences 
	// in successive not chosen 'or' sequence
	stepsToBeDone=stepNumber-i;
   
	if(StartVEFrame.verbose)
	    System.out.println
	       ("stepsToBeDone  " + stepsToBeDone+ " in order #"+ orderNumber);

	return stepsToBeDone;
  }

  /**
   * returning the production phase at a specific step
   */

    public int getProductionToBeDoneAtStepAheadNumber (int sn) {

	int i, p;
	int step;

	step = sn;
	for (i=0;i< stepNumber && orderState[i]!=0;i++); 
	                          // to find the first position
	                          // where orderState[i] == 0

	numberOfTheStepForWhichANewsIsSent=i+step-1;
	p=getOrderRecipe(numberOfTheStepForWhichANewsIsSent);
	if(p>=1000000000)p-=1000000000;

	return p;
    }

 /**
   * returning the sequential number of the recipe referring to which
   * a news will be sent
   */
    public int getNumberOfTheStepForWhichANewsIsSent () {

	return numberOfTheStepForWhichANewsIsSent;
  }

  /** returning the value of the production already done */
    public float getOrderEvaluation(){
	int globalOrderState = 0;
	float value = 0;

	// evaluate the number of steps done
	for(int i=0; i<orderState.length; i++){
	    if(orderState[i]>0)globalOrderState++;
	}

	value = globalOrderState * 
	    vEFrameModelSwarm.getRevenuePerEachRecipeStep();
	return value;
    }

  /** setting fixed and variable costs for the orders in production
   *  or finished */
    public void setFixedAndVariableOrderCosts(float fc, float vc){

	cumulatedFixedCost += fc;
	cumulatedVariableCost += vc;
	orderTotalCosts = (cumulatedFixedCost + cumulatedVariableCost);
  }

  /** getting the total of costs of the order */ 
    public float getOrderTotalCosts(){
	return orderTotalCosts;
    }

    /** getting the content of orderRecipe[i] */
    public int getOrderRecipe(int i){
	if(i>=stepNumber){
	    System.out.println(
	 "Error in getting a recipe step in an order (max length exceeded)");
	    MyExit.exit(1);
	}
	return orderRecipe[i];
    }

    /** getting the content of orderState[i] */
    public int getOrderState(int i){
	if(i>=stepNumber){
	    System.out.println(
	 "Error in getting a state value in an order (max length exceeded)");
	    MyExit.exit(1);
	}
	return orderState[i];
    }

    /** getting the procurement specification set related to the nextStep
     *  (the current one, if the order is in a waiting list for next step
     *  to be accomplished)<br>
     *  if no procurement specification set exists for the next step,
     *  a null value is returned */
    public Object getPendingProcurementSpecificationSet()
    {
	int i, position;
	ProcurementSpecificationSet pendingProcurementSpecificationSet;

	position=-1;
	for (i=0;i< stepNumber && position==-1;i++)
	    if (orderState[i]==0)position=i;

	pendingProcurementSpecificationSet = null;

	if(procurementSpecificationSetList.getCount()>0){
	    procurementSpecificationSetListIndex.setOffset(0);
	    for(i=0;i<procurementSpecificationSetList.getCount();i++){
		if(position == ((ProcurementSpecificationSet)
				procurementSpecificationSetListIndex.get()).
		   getPositionInRecipe())
		    pendingProcurementSpecificationSet =
			(ProcurementSpecificationSet)
			procurementSpecificationSetList.atOffset(i);
		procurementSpecificationSetListIndex.next();
	    }
	}
	return pendingProcurementSpecificationSet;
    }

    /** getting the computational specification set related to the nextStep
     *  (the current one, if the order is in a waiting list for next step
     *  to be accomplished)<br>
     *  if no computational specification set exists for the next step,
     *  a null value is returned */
    public Object getPendingComputationalSpecificationSet()
    {
	int i, position;
	ComputationalSpecificationSet pendingComputationalSpecificationSet;

	position=-1;
	for (i=0;i< stepNumber && position==-1;i++)
	    if (orderState[i]==0)position=i;

	pendingComputationalSpecificationSet = null;

	if(computationalSpecificationSetList.getCount()>0){
	    computationalSpecificationSetListIndex.setOffset(0);
	    for(i=0;i<computationalSpecificationSetList.getCount();i++){
		if(position == ((ComputationalSpecificationSet)
				computationalSpecificationSetListIndex.get()).
		   getPositionInRecipe())
		    pendingComputationalSpecificationSet =
			(ComputationalSpecificationSet)
			computationalSpecificationSetList.atOffset(i);
		computationalSpecificationSetListIndex.next();
	    }
	}
	return pendingComputationalSpecificationSet;
    }


    /** getting the stand alone batch specification set related to the nextStep
     *  (the current one, if the order is in a waiting list for next step
     *  to be accomplished)<br>
     *  if no stand alone batch SpecificationSet exists for the next step,
     *  a null value is returned */
    public Object getPendingStandAloneBatchSpecificationSet()
    {
	int i, position;
	StandAloneBatchSpecificationSet pendingStandAloneBatchSpecificationSet;

	position=-1;
	for (i=0;i< stepNumber && position==-1;i++)
	    if (orderState[i]==0)position=i;

	pendingStandAloneBatchSpecificationSet = null;

	if(standAloneBatchSpecificationSetList.getCount()>0){
	    standAloneBatchSpecificationSetListIndex.setOffset(0);
	    for(i=0;i<standAloneBatchSpecificationSetList.getCount();i++){
		if(position == ((StandAloneBatchSpecificationSet)
			    standAloneBatchSpecificationSetListIndex.get()).
		   getPositionInRecipe())
		    pendingStandAloneBatchSpecificationSet =
			(StandAloneBatchSpecificationSet)
			standAloneBatchSpecificationSetListIndex.get();
		standAloneBatchSpecificationSetListIndex.next();
	    }
	}
	return pendingStandAloneBatchSpecificationSet;
    }

    /** getting the sequential batch specification set related to the nextStep
     *  (the current one, if the order is in a waiting list for next step
     *  to be accomplished)<br>
     *  if no sequential batch SpecificationSet exists for the next step,
     *  a null value is returned */
    public Object getPendingSequentialBatchSpecificationSet()
    {
	int i, position;
	SequentialBatchSpecificationSet pendingSequentialBatchSpecificationSet;

	position=-1;
	for (i=0;i< stepNumber && position==-1;i++)
	    if (orderState[i]==0)position=i;

	pendingSequentialBatchSpecificationSet = null;

	if(sequentialBatchSpecificationSetList.getCount()>0){
	    sequentialBatchSpecificationSetListIndex.setOffset(0);
	    for(i=0;i<sequentialBatchSpecificationSetList.getCount();i++){
		if(position == ((SequentialBatchSpecificationSet)
			    sequentialBatchSpecificationSetListIndex.get()).
		   getPositionInRecipe())
		    pendingSequentialBatchSpecificationSet =
			(SequentialBatchSpecificationSet)
			sequentialBatchSpecificationSetListIndex.get();
		sequentialBatchSpecificationSetListIndex.next();
	    }
	}
	return pendingSequentialBatchSpecificationSet;
    }

    /** getting the 'or' specification set related to the nextStep
     *  (the current one, if the order is in a waiting list for next step
     *  to be accomplished)<br>
     *  if no 'or' SpecificationSet exists for the next step,
     *  a null value is returned */
    public Object getPendingOrSpecificationSet()
    {
	int i, position;
	OrSpecificationSet pendingOrSpecificationSet;

	position=-1;
	for (i=0;i< stepNumber && position==-1;i++)
	    if (orderState[i]==0)position=i;

	pendingOrSpecificationSet = null;

	if(orSpecificationSetList.getCount()>0){
	    orSpecificationSetListIndex.setOffset(0);
	    for(i=0;i<orSpecificationSetList.getCount();i++){
		if(position == ((OrSpecificationSet) 
				orSpecificationSetListIndex.get()).
		   getPositionInRecipe())
		    pendingOrSpecificationSet =
			(OrSpecificationSet) orSpecificationSetListIndex.get();
		orSpecificationSetListIndex.next();
	    }
	}
	return pendingOrSpecificationSet;
    }

    /** adding an item to the procuredItemList */
    public void setProcuredItemList(Order o){
	procuredItemList.addLast(o);
    }

    /** set multiplicity factor */
    public void setMultiplicity(int m){
	multiplicity = m;
    }

    /** get multiplicity factor */
    public int getMultiplicity(){
	return multiplicity;
    }


    /** drop the order */
    public void drop(){
	Order aProcuredItem;
	StandAloneBatchSpecificationSet tmp;

	// dropping our procurementSpecificationSet
	while(procurementSpecificationSetList.getCount()>0)
	    ((ProcurementSpecificationSet) 
	     procurementSpecificationSetList.removeFirst()).drop();

	// dropping procured items (which in turn will be dropping their
	// procurementSpecificationSet, their procured items and their
	// standAloneBatchSpecificationSet + sequentialBatchSpecificationSet
	while(procuredItemList.getCount()>0)
	    {
		aProcuredItem = (Order) procuredItemList.removeFirst();
		((ListImpl) vEFrameModelSwarm.getOrderList()).
		    remove(aProcuredItem); 
		aProcuredItem.drop();
	    }

	// dropping our computationalSpecificationSet
	while(computationalSpecificationSetList.getCount()>0)
	    ((ComputationalSpecificationSet) 
	     computationalSpecificationSetList.removeFirst()).drop();

	// dropping all our standAloneBatchSpecificationSet
	while(standAloneBatchSpecificationSetList.getCount()>0)
	    {
		tmp = (StandAloneBatchSpecificationSet) 
		       standAloneBatchSpecificationSetList.removeFirst();
		// the specification set at the address copied in tmp
		// may exist no more, having been dropped in the cleaning
		// operation of a clone of the order we deal with
		if(checkMethod("tmp","getNumber") != null)tmp.drop();
	    }

	// dropping all our sequentialBatchSpecificationSet (without
	// the cloning problem reported above)
	while(sequentialBatchSpecificationSetList.getCount()>0)
	    ((SequentialBatchSpecificationSet) 
	     sequentialBatchSpecificationSetList.removeFirst()).
		drop();

	// dropping all our orSpecificationSet (without
	// the cloning problem reported above)
	while(orSpecificationSetList.getCount()>0)
	    ((OrSpecificationSet) orSpecificationSetList.removeFirst()).drop();

	super.drop();
}

    public Selector checkMethod(Object obj, String method)
    {
	Selector sel;

	try
	    {
	    sel = new Selector(obj.getClass(), method, false);
	    } catch (Exception noNameUsed)
		{
		    return null;
		}

	return sel;
    }


    /** settinging the name of the recipe */
    public void setRecipeName(String s){
	recipeName=s;
    }

    /** retrieving the name of the recipe */
    public String getRecipeName(){
	return recipeName;
    }


    /** getting the order recipe vector address */
    public int [] getRecipeVector(){
	return orderRecipe;
    }

    /** getting the order state vector address */
    public int [] getStateVector(){
	return orderState;
    }

    /** set inSequentialBatch */
    public void setInSequentialBatch(boolean v){
	inSequentialBatch=v;
    }

    /** get inSequentialBatch */
    public boolean getInSequentialBatch(){
	return inSequentialBatch;
    }

    /**
     * Cloneable interface method: clone() 
     */
    public Object clone()
    {
	Order orderClone;

	// the cloning operation, via the specialized constr
	orderClone = new Order(getZone(), orderRecipe.length, 
			       orderRecipe, orderState.length, orderState,
			       vEFrameModelSwarm,
			       procurementSpecificationSetListIndex,
			       procuredItemListIndex,
			       computationalSpecificationSetListIndex,
			       standAloneBatchSpecificationSetListIndex,
			       sequentialBatchSpecificationSetListIndex,
			       orSpecificationSetListIndex,
			       inSequentialBatch);
	orderClone.setMultiplicity(1);

        // add the active orders to the general order list (they will be
	// eliminated when dropped in a unit, being finished); this
	// list has been introduced for accounting purposes
 	((ListImpl) vEFrameModelSwarm.getOrderList()).
	    addLast(orderClone); 

	return orderClone;
    }

//    /**
//     * Bitwise cloneable interface method: clone() 
//     */     
//    public Object clone()
//    {
//	return super.super.clone();
//    }


    /** the internal method dealing with procurements */
    public void arrangeForProcurements()	
    {
	jj=j;
	j += r[j+1] + 2;
	if (! (j<stepNumber)){System.out.
	println(
	"Internal dimension error analysing the order: " + 
	"look at the constructor of Order class.");
	MyExit.exit(1);
	}
	
	// making a procurementSpecificationSet object, linked
	// at the position 'i' of the recipe
	aProcurementSpecificationSet = new 
	    ProcurementSpecificationSet(getZone(), 
					orderNumber, i);
	
	procurementSpecificationSetList.
	    addLast(aProcurementSpecificationSet);
	
	// setting specifications into 
	// aProcurementSpecificazioneSet
	aProcurementSpecificationSet.
	    setSpecificationSet(jj,r);
	
	// being after a "p" specification, stop execution
	// if the unit is an endUnit
	for(k=0;k<endUnitList.getCount();k++)
	    if(r[j]==((EndUnit) endUnitList.atOffset(k)).
	       getUnitNumber()){
		System.out.println(
			   "Syntax error in Order # " + getOrderNumber());
		System.out.println("An endUnit cannot follow " + 
			   "immediately after a procurement step");
		MyExit.exit(1);
	    }

	// copying the recipe
	orderRecipe[i] = r[j];
	
    }

    /** the internal method dealing with computation */
    public void arrangeForComputation()	
    {

	jj=j;

	// r[jj]   contains the computational code
	// r[jj+1] contains the number of matrixes used
	j += r[j+1] + 2;
	if (! (j<stepNumber)){System.out.
	println(
	"Internal dimension error analysing the order: " + 
	"look at arrangeForComputation() method of Order class.");
	MyExit.exit(1);
	}

	// making a ComputationalSpecificationSet object, linked
	// at the position 'i' of the recipe
	aComputationalSpecificationSet = new 
	    ComputationalSpecificationSet(getZone(), this, i);
	
	computationalSpecificationSetList.
	    addLast(aComputationalSpecificationSet);

	// building the vector of memory matrix addresses
	for (int m=0; m<r[jj+1];m++)
	    matrixAddress[m]=vEFrameModelSwarm.getMemoryMatrix(r[jj+2+m]);
	
	// setting specifications into 
	// aComputatatioanlSpecificazioneSet
	aComputationalSpecificationSet.
	    setSpecificationSet(r[jj+1], matrixAddress, r[jj]);
	
	// being after a "c" specification, stop execution
	// if the unit is an endUnit
	for(k=0;k<endUnitList.getCount();k++)
	    if(r[j]==((EndUnit) endUnitList.atOffset(k)).
	       getUnitNumber()){
		System.out.println(
			   "Syntax error in Order # " + getOrderNumber());
		System.out.println("An endUnit cannot follow " + 
			   "immediately after a computational step");
		MyExit.exit(1);
	    }

	// copying the recipe
	orderRecipe[i] = r[j];
	
    }
}
